@@ -1,7 +1,7 @@
 /*
  * main.c - Unix Teredo server & relay implementation
  *          command line handling and core functions
- * $Id: main.c 2052 2007-10-03 18:53:24Z remi $
+ * $Id: main.c 2093 2008-01-05 14:37:52Z remi $
  */
 
 /***********************************************************************
@@ -153,21 +153,29 @@ error_errno (const char *str)
 # define O_NOFOLLOW 0
 #endif
 static int
-open_pidfile (const char *path)
+create_pidfile (const char *path)
 {
 	int fd;
 
 	fd = open (path, O_WRONLY|O_CREAT|O_NOFOLLOW, 0644);
 	if (fd != -1)
 	{
+		char buf[20]; // enough for > 2^64
 		struct stat s;
 
+		snprintf (buf, sizeof (buf), "%d", (int)getpid ());
+		buf[sizeof (buf) - 1] = '\0';
+
+		fcntl (fd, F_SETFD, fcntl (fd, F_GETFD) | FD_CLOEXEC);
 		errno = 0;
-		/* We only check the lock. The actual locking occurs
-		 * after (possibly) calling daemon(). */
+
+		/* Locks the PID file */
 		if ((fstat (fd, &s) == 0)
-		 && S_ISREG(s.st_mode)
-		 && (lockf (fd, F_TEST, 0) == 0))
+		 && S_ISREG (s.st_mode)
+		 && (lockf (fd, F_TLOCK, 0) == 0)
+		 && (ftruncate (fd, 0) == 0)
+		 && (write (fd, buf, strlen (buf)) == (ssize_t)strlen (buf))
+		 && (fdatasync (fd) == 0))
 			return fd;
 
 		if (errno == 0) /* !S_ISREG */
@@ -175,42 +183,11 @@ open_pidfile (const char *path)
 
 		(void)close (fd);
 	}
+
 	return -1;
 }
 
 
-static int
-write_pid (int fd)
-{
-	char buf[20]; // enough for > 2^64
-
-	/* Actually lock the file */
-	if (lockf (fd, F_TLOCK, 0))
-		return -1;
-
-	(void)snprintf (buf, sizeof (buf), "%d", (int)getpid ());
-	buf[sizeof (buf) - 1] = '\0';
-	size_t len = strlen (buf);
-
-	if (ftruncate (fd, 0)
-	 || (write (fd, buf, len) != (ssize_t)len)
-	 || fdatasync (fd))
-		return -1;
-	return 0;
-}
-
-
-static int
-close_pidfile (int fd)
-{
-	if (lockf (fd, F_ULOCK, 0))
-		return -1;
-	if (close (fd))
-		return -1;
-	return 0;
-}
-
-
 #ifdef MIREDO_DEFAULT_USERNAME
 static void
 setuid_notice (void)
@@ -231,8 +208,9 @@ init_security (const char *username)
 {
 	int val;
 
-	/* Sets sensible umask */
 	(void)umask (022);
+	if (chdir ("/"))
+		return -1;
 
 	/*
 	 * We close all file handles, except 0, 1 and 2.
@@ -491,8 +469,38 @@ int miredo_main (int argc, char *argv[])
 	if (miredo_diagnose ())
 		return 1;
 
+	int pipes[2];
+	if (pipe (pipes))
+		pipes[0] = pipes[1] = -1;
+
+	if (!flags.foreground)
+	{
+		pid_t pid = fork ();
+
+		switch (pid)
+		{
+			case -1:
+				fprintf (stderr, _("Error (%s): %s\n"), "fork", strerror (errno));
+				return 1;
+
+			case 0:
+				break;
+
+			default:
+			{
+				close (pipes[1]);
+				if (read (pipes[0], &c, sizeof (c)) != sizeof (c))
+					c = 1;
+
+				close (pipes[0]);
+				return c;
+			}
+		}
+	}
+	close (pipes[0]);
+
 	/* Opens pidfile */
-	int fd = open_pidfile (pidfile);
+	int fd = create_pidfile (pidfile);
 	if (fd == -1)
 	{
 		fprintf (stderr, _("Cannot create PID file %s:\n %s\n"),
@@ -501,30 +509,31 @@ int miredo_main (int argc, char *argv[])
 			fprintf (stderr, "%s\n",
 			         _("Please make sure another instance of the program is "
 				   "not already running."));
-		return -1;
+		exit (1);
 	}
 
 	/* Detaches */
-	if (!flags.foreground && daemon (0, 0))
+	if (!flags.foreground)
 	{
-		fprintf (stderr, _("Error (%s): %s\n"), "daemon", strerror (errno));
-		return -1;
-	}
+		c = 0;
 
-	if (write_pid (fd))
-	{
-		close (fd);
-		return -1;
+		setsid ();
+		if (freopen ("/dev/null", "r", stdin) == NULL
+		 || freopen ("/dev/null", "w", stdout) == NULL
+		 || freopen ("/dev/null", "w", stderr) == NULL
+		 || (write (pipes[1], &c, sizeof (c)) <= 0))
+			exit (1);
 	}
+	close (pipes[1]);
 
 	/*
 	 * Run
 	 */
 	c = miredo (conffile, servername, fd);
 
-	close_pidfile (fd);
 	(void)unlink (pidfile);
+	close (fd);
 
-	return c ? 1 : 0;
+	exit (c ? 1 : 0);
 }
 
